#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <vector>
#include <algorithm>
#include <string>
#include <math.h>
#include<sys/stat.h>
#include<sys/types.h>
#ifdef WINDOWS
#include <direct.h>
#define GetCurrentDir _getcwd
#else
#include <unistd.h>
#define GetCurrentDir getcwd
#endif
#include "SpiceUsr.h"

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

//#if CONFIG_USE_DOUBLE
//
//#if defined(cl_khr_fp64)  // Khronos extension available?
//#pragma OPENCL EXTENSION cl_khr_fp64 : enable
////#define DOUBLE_SUPPORT_AVAILABLE
//#elif defined(cl_amd_fp64)  // AMD extension available?
//#pragma OPENCL EXTENSION cl_amd_fp64 : enable
////#define DOUBLE_SUPPORT_AVAILABLE
//#endif
//
//#endif // CONFIG_USE_DOUBLE

#ifdef cl_khr_fp64
#pragma OPENCL EXTENSION cl_khr_fp64 : enable
#elif defined(cl_amd_fp64)
#pragma OPENCL EXTENSION cl_amd_fp64 : enable
#endif

#define UTCFMT "YYYY-MON-DD-HR:MN:SC ::UTC"
#define TIMLEN 51
#define PI 3.14159265358979323846

using namespace std;
void err_check(int err, string err_code) {
	if (err != CL_SUCCESS) {
		cout << "Error: " << err_code << "(" << err << ")" << endl;
		exit(-1);
	}
}

double interp(double x_0, double x_1, double y_0, double y_1, double x) {
	double y;

	y = y_0 + (y_1 - y_0) * ((x - x_0) / (x_1 - x_0));
	return y;
}

int main(int argc, char** argv) {

	//-----	OPENCL VARIABLE DECLARATION ---
	cl_program program = NULL;
	cl_kernel kernel = NULL;
	cl_platform_id platform_id = NULL;
	cl_uint ret_num_platforms;
	cl_uint ret_num_devices;
	cl_device_id device_id = NULL;
	cl_context context = NULL;
	cl_command_queue command_queue = NULL;

	cl_mem mobj_elev_max = 0;
	cl_mem mobj_elev_min = 0;
	cl_mem mobj_az_file = 0;

	cl_mem mobj_x_vec = 0;
	cl_mem mobj_y_vec = 0;
	cl_mem mobj_z_vec = 0;

	cl_mem mobj_sun = 0;
	cl_mem mobj_out = 0;

	cl_int errNum;

	cout << "---------------------------------" << endl;
	cout << "-------  TUB-Illu OpenCL  -------" << endl;
	cout << "------    Version: 1.0     ------" << endl;
	cout << "---------------------------------\n" << endl;

	cout << "Setting parameters..." << endl;
	char startUTC[25];
	char stopUTC[25];
	//char startUTC[] = "2022-FEB-01-00:00:00";
	//char stopUTC[] = "2023-FEB-01-00:00:00";
	//char startUTC[] = "2015 DEC 01 00:00:00";
	//char stopUTC[] = "2034 DEC 01 00:00:00";
	//char startUTC[] = "2018 OCT 22 00:00:00";
	//char stopUTC[] = "2019 OCT 22 00:00:00";
	//char stopUTC[] = "2018 NOV 22 00:00:00";
	SpiceDouble etbeg, etend;
	double pos_sun[3];

	ofstream out_file, sorted, sun_spice_fp;
	//sun_spice_fp.open("/users/figu_ra/workspace/Illu_CL/phil_sun_ele_spice.txt");

	double pos_cart_in[3];
	double pos_sun_short[3];
	double pos_obs_sun[3];

	double lt;
	double stepsz = 3600; //in seconds

	double p;
	double c;
	double radius = 1737.4;
	double lon0 = 0;
	double lat0 = -90;
	double rho = PI / 180.0;
	double lat, lon;
	double input[3] = { -11.55, -12.47, 1.95879 };
	//double input[3] = { 122.85, 185.59, -1.3678 };


	double angle;
	double r, longitude, latitude;
	double elev, k, cos_c, x_sun, y_sun;

	//vector<double> eft;
	float fil = 0;
	int read_count;
	float az_in_vec[721];
	ifstream readin[720], get_z;
	float azimuth = 0;
	int size_roi;
	int lqcip;
	//vector<double> elev_in_vec_min, elev_in_vec_max;


	ofstream sunsun_fp;
	//position sun and moon

	ifstream read_in;
	string line;
	string direct;
	string path;
	int flag_data, flag_tmstmp, config;

	read_in.open(argv[1], ios::in);
	if (!read_in) {
		cerr << " File " << "config_illumination.txt couldn't be opened! "
				<< endl;
	}

	for (int i = 0; i < 5; i++)
		getline(read_in, line);
	// PATH TO LIST OF LOLA TRACKS AND THE DIRECTORY WHERE RESULTS SHOULD BE STORED
	read_in >> startUTC;
	getline(read_in, line);
	read_in >> stopUTC;
	getline(read_in, line);
	read_in >> size_roi;
	getline(read_in, line);
	read_in >> flag_tmstmp;
	getline(read_in, line);
	read_in >> flag_data;
	getline(read_in, line);
	read_in >> direct;
	getline(read_in, line);
	read_in >> path;
	getline(read_in, line);
	read_in >> lqcip;
	getline(read_in, line);
	read_in >> config;
	getline(read_in, line);

	read_in.close();
	if (config == 0){
	cout << "Floating Point configuration selected" << endl;
	}else if (config ==1){
		cout << "Double Precision configuration selected" << endl;
	}

	float *elev_in_vec_min = new float[size_roi];
	float *elev_in_vec_max = new float[size_roi];
	//vector<float> x_vec, y_vec, z_vec;
	float *x_vec = new float[size_roi];
	float *y_vec = new float[size_roi];
	float *z_vec = new float[size_roi];

	float x, y, z, ele2;

	float id_min, id_max;

	id_min = 0;
	float az_file[1];
	//float visible[size_roi];
	float *visible = new float[size_roi];
	int i_count = 0;
	//float accum_temp[size_roi];
	float *accum_temp = new float[size_roi];

	cout << "Reading kernels..." << endl;
	furnsh_c("/sys_shared/kernels/common/lsk/naif0010.tls");
	furnsh_c("/sys_shared/kernels/common/pck/moon_pa_de421_1900-2050.bpc");
	furnsh_c("/sys_shared/kernels/lro/fk/moon_080317.tf");
	furnsh_c("/sys_shared/kernels/common/spk/planets/de421.bsp");
	furnsh_c("/sys_shared/kernels/lro/fk/moon_assoc_me.tf");
	furnsh_c("/sys_shared/kernels/common/pck/pck00010.tpc");

	utc2et_c(startUTC, &etbeg);
	utc2et_c(stopUTC, &etend);
	int out = 0;

	//vector<double> az_vec, order_vec, sun_x_v, sun_y_v, sun_z_v;
	cout << "Calculating position of the sun from " << startUTC << " to "
			<< stopUTC << endl;
	//out_file.open("/users/figu_ra/workspace/Illu_CL/opencl/time_pos.txt");
	int filler = ((etend - etbeg) / 3600) + 1;
	//cout << filler << endl;

	float *az_vec = new float[filler];
	float *order_vec = new float[filler];
	float *sun_x_v = new float[filler];
	float *sun_y_v = new float[filler];
	float *sun_z_v = new float[filler];
	double *eft = new double[filler];

	for (double i = etbeg; i <= etend; i = i + stepsz) {

		spkpos_c("sun", i, "MOON_ME", "LT", "MOON", pos_sun, &lt); //position of sun wrt moon

		//--------------------------------


		//------------------------------------

		p = sqrt(input[0] * input[0] + input[1] * input[1]);
		c = atan(p / radius);
		lon = lon0 + atan2(input[0] * sin(c), (p * cos(lat0 * rho) * cos(c)
				- input[1] * sin(lat0 * rho) * sin(c))) / rho;
		lat = asin(cos(c) * sin(lat0 * rho) + input[1] * sin(c) * cos(lat0
				* rho) / p) / rho;
		if (lon < 0)
			lon += 360.0;

		//---------------------------------------

		latrec_c(1737.4 + input[2], lon / dpr_c(), lat / dpr_c(), pos_cart_in);

		vsub_c(pos_sun, pos_cart_in, pos_obs_sun);

		elev = 90 - (vsep_c( pos_cart_in,pos_obs_sun) * dpr_c());

		pos_sun_short[0] = pos_cart_in[0] + pos_obs_sun[0] * (200
				/ vnorm_c(pos_obs_sun));
		pos_sun_short[1] = pos_cart_in[1] + pos_obs_sun[1] * (200
				/ vnorm_c(pos_obs_sun));
		pos_sun_short[2] = pos_cart_in[2] + pos_obs_sun[2] * (200
				/ vnorm_c(pos_obs_sun));

		reclat_c(pos_sun_short, &r, &longitude, &latitude);

		cos_c = sin(lat0 * rho) * sin(latitude) + cos(lat0 * rho) * cos(
				latitude) * cos(longitude - lon0 * rho);
		k = 1.0 / cos_c;
		x_sun = radius * k * cos(latitude) * sin(longitude - lon0 * rho);
		y_sun = radius * k * (cos(lat0 * rho) * sin(latitude) - sin(lat0 * rho)
				* cos(latitude) * cos(longitude - lon0 * rho));

		angle = atan2((x_sun - input[0]), (y_sun - input[1])) * 180 / PI;//

		if (angle < 0) {
			angle = angle + 360;
		}

		if (angle > 360) {
			cout << "out of bounds " << fixed << i << endl;
		}
		//sun_spice_fp << fixed << i << " " << elev << " " << angle << endl;
		//sun_spice_fp << out  << " "<< angle << " " << elev << endl;
		//out_file << fixed << out << " " << angle << " " << pos_obs_sun[0]
		//<< " " << pos_obs_sun[1] << " " << pos_obs_sun[2] << endl;

		sun_x_v[out] = pos_obs_sun[0];
		sun_y_v[out] = pos_obs_sun[1];
		sun_z_v[out] = pos_obs_sun[2];
		order_vec[out] = out;
		az_vec[out] = angle;
		eft[out] = i;

		out++;
	}
	//cout << fixed << "Distance to sun: " << vnorm_c(pos_sun) << endl;
	cout << out << " hours calculated." << endl;
	cout << " Calculation finished..." << endl;

	//--------SORTING BY AZIMUTH
	cout << " Sorting the positions of the sun azimuth wise..." << endl;

	double t, ord, elesunx, elesuny, elesunz, ettime;

	int i, j;
	for (i = 1; i <= out; i++)
		for (j = 1; j <= out - i; j++)
			if (az_vec[j - 1] >= az_vec[j]) {
				t = az_vec[j - 1];
				ord = order_vec[j - 1];
				elesunx = sun_x_v[j - 1];
				elesuny = sun_y_v[j - 1];
				elesunz = sun_z_v[j - 1];
				ettime = eft[j - 1];

				az_vec[j - 1] = az_vec[j];
				order_vec[j - 1] = order_vec[j];
				sun_x_v[j - 1] = sun_x_v[j];
				sun_y_v[j - 1] = sun_y_v[j];
				sun_z_v[j - 1] = sun_z_v[j];
				eft[j - 1] = eft[j];

				az_vec[j] = t;
				order_vec[j] = ord;
				sun_x_v[j] = elesunx;
				sun_y_v[j] = elesuny;
				sun_z_v[j] = elesunz;
				eft[j] = ettime;
			}

	cout << " Sun azimuths sorted..." << endl;
	//--------INTERPOLATING AND CHECKING IF ILLUMINATED OR NOT
	const char *az_path_zero;

	cout << " Calculating the illumination..." << endl;

	for (int i = 0; i <= 720; i++) {
		az_in_vec[i] = fil;
		fil = fil + 0.5;
	}

	for (int i_az = 0; i_az < 720; i_az++) {

		stringstream ss;
		string str;
		if (flag_data == 1) {
			if (i_az % 2 == 0.0)
				ss << direct << "azimuth_" << azimuth << ".0.txt";
			else
				ss << direct << "azimuth_" << azimuth << ".txt";
		} else if (flag_data == 0) {
			ss << direct << "azimuth_" << azimuth << ".txt";
		}
		ss >> str;

		if (i_az == 0) {
			az_path_zero = str.c_str();
		}

		readin[i_az].open(str.data());

		if (!readin[i_az].is_open())
			cout << "Unable to open file\n";

		azimuth = azimuth + 0.5;

	}

	if (flag_tmstmp == 1) {
		if (mkdir("/users/figu_ra/workspace/Illu_CL/results/visible_files", 0777)
				== -1)//creating a directory
		{
			cerr << "Error no directory created" << endl;
			return (1);
		}

	}
	//	if (mkdir("/users/figu_ra/workspace/Illu_CL/visible_files", 0777) == -1)//creating a directory
	//	{
	//		cerr << "Error no directory created" << endl;
	//		return (1);
	//	}


	az_file[0] = az_vec[i_count];
	float sun_file[5];
	sun_file[0] = sun_x_v[i_count];
	sun_file[1] = sun_y_v[i_count];
	sun_file[2] = sun_z_v[i_count];

	int off = 0;

	for (int i = 0; i < size_roi; i++) {
		accum_temp[i] = 0;
	}

	///-----GETTING THE HEIGHT OF EACH POINT

	const char *roi_path = path.c_str();

	get_z.open(roi_path);

	double nyu;
	for (int a = 0; a < size_roi; a++) {
		if (flag_data == 1) {
			get_z >> x >> y >> z >> nyu;
		}
		if (flag_data == 0) {
			get_z >> x >> y >> z;
		}

		x_vec[a] = x;
		y_vec[a] = y;
		z_vec[a] = z;

	}
	get_z.close();

	sunsun_fp.open("/users/figu_ra/workspace/Illu_CL/results/sunsun.txt");

	//-----SET UP THE OPENCL ENVIRONMENT
	clock_t time = clock();
	// return code used by OpenCL API
	cl_int status;

	// wait event synchronization handle used by OpenCL API
	cl_event event;

	//step 01: Get platform/device information
	errNum = clGetPlatformIDs(1, &platform_id, &ret_num_platforms);
	err_check(errNum, "clGetPlatfromIDs");

	//step 02: Get information about device
	errNum = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, 1, &device_id,
			&ret_num_devices);
	err_check(errNum, "clGetDeviceIDs");

	//step 03: Create OpenCL Context
	context = clCreateContext(NULL, 1, &device_id, NULL, NULL, &errNum);
	err_check(errNum, "clCreateContext");

	//step 04: Create Command Queue

	command_queue = clCreateCommandQueue(context, device_id, 0, &errNum);
	err_check(errNum, "clCreateCommandQueue");

	if (config == 0) {
		//step 05: Read kernel file
		ifstream file("./kernel_float.cl");
		string prog(istreambuf_iterator<char> (file),
				(istreambuf_iterator<char> ()));
		const char *source_str = prog.c_str();
		// Step 06: Create Kernel program from the read in source
		program = clCreateProgramWithSource(context, 1,
				(const char **) &source_str, 0, &errNum);
		err_check(errNum, "clCreateProgramWithSource");

	} else if (config == 1) {
		//step 05: Read kernel file
		ifstream file("./kernel_double.cl");
		string prog(istreambuf_iterator<char> (file),
				(istreambuf_iterator<char> ()));
		const char *source_str = prog.c_str();
		// Step 06: Create Kernel program from the read in source
		program = clCreateProgramWithSource(context, 1,
				(const char **) &source_str, 0, &errNum);
		err_check(errNum, "clCreateProgramWithSource");

	}
	//	// Step 06: Create Kernel program from the read in source
	//	program = clCreateProgramWithSource(context, 1,
	//			(const char **) &source_str, 0, &errNum);
	//	err_check(errNum, "clCreateProgramWithSource");

	// Step 07: Build Kernel Program
	errNum = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
	//err_check(errNum, "clBuildProgram");

	//Determine the size of the log
	size_t log_size;
	clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, 0, NULL,
			&log_size);

	// Allocate memory for the log
	char *log = (char *) malloc(log_size);

	// Get the log
	clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, log_size,
			log, NULL);

	// Print the log
	printf("%s\n", log);

	// Step 08: Create OpenCL Kernel
	kernel = clCreateKernel(program, "illumination", &errNum);
	err_check(errNum, "clCreateKernel");

	// Step 09: Create memory objects and transfer the data to memory buffer

	mobj_x_vec = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float)
			* size_roi, NULL, NULL);
	errNum = clEnqueueWriteBuffer(command_queue, mobj_x_vec, CL_TRUE, 0,
			sizeof(float) * size_roi, x_vec, 0, NULL, &event);
	status = clWaitForEvents(1, &event);
	clReleaseEvent(event);
	err_check(errNum, "clSetMemObj");

	mobj_y_vec = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float)
			* size_roi, NULL, NULL);
	errNum = clEnqueueWriteBuffer(command_queue, mobj_y_vec, CL_TRUE, 0,
			sizeof(float) * size_roi, y_vec, 0, NULL, &event);
	status = clWaitForEvents(1, &event);
	clReleaseEvent(event);
	err_check(errNum, "clSetMemObj");

	mobj_z_vec = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float)
			* size_roi, NULL, NULL);
	errNum = clEnqueueWriteBuffer(command_queue, mobj_z_vec, CL_TRUE, 0,
			sizeof(float) * size_roi, z_vec, 0, NULL, &event);
	status = clWaitForEvents(1, &event);
	clReleaseEvent(event);
	err_check(errNum, "clSetMemObj");

	//---SET UP SIZE OF GLOBAL AND LOCAL WORK

	size_t globalWorkSize[1] = { size_roi };
	size_t localWorkSize[1] = { 250 };

	//------START THE ILLUMINATION
	sun_file[3] = id_min;
	for (read_count = 0; read_count < 720; read_count++) {

		if (read_count == 719) {
			readin[0].open(az_path_zero);
		}
		for (int a = 0; a < size_roi; a++)//while not end of file
		{
			if (flag_data == 1) {
				readin[read_count] >> ele2 >> y;
			}
			if (flag_data == 0) {
				readin[read_count] >> x >> y >> ele2;
			}
			elev_in_vec_min[a] = ele2;

			if (read_count == 719) {
				if (flag_data == 1) {
					readin[0] >> ele2 >> y;
				}
				if (flag_data == 0) {
					readin[0] >> x >> y >> ele2;
				}
				elev_in_vec_max[a] = ele2;
			} else {
				if (flag_data == 1) {
					readin[read_count + 1] >> ele2 >> y;
				}
				if (flag_data == 0) {
					readin[read_count + 1] >> x >> y >> ele2;
				}
				elev_in_vec_max[a] = ele2;

			}
		}
		if (read_count < 719) {
			readin[read_count + 1].clear();
			readin[read_count + 1].seekg(0, ios::beg);
		}
		readin[read_count].close();
		id_max = az_in_vec[read_count + 1];
		sun_file[4] = id_max;

		mobj_elev_min = clCreateBuffer(context, CL_MEM_READ_WRITE,
				sizeof(float) * size_roi, NULL, NULL);
		errNum = clEnqueueWriteBuffer(command_queue, mobj_elev_min, CL_TRUE, 0,
				sizeof(float) * size_roi, elev_in_vec_min, 0, NULL, &event);
		status = clWaitForEvents(1, &event);
		clReleaseEvent(event);
		err_check(errNum, "clSetMemObj");

		mobj_elev_max = clCreateBuffer(context, CL_MEM_READ_WRITE,
				sizeof(float) * size_roi, NULL, NULL);
		errNum = clEnqueueWriteBuffer(command_queue, mobj_elev_max, CL_TRUE, 0,
				sizeof(float) * size_roi, elev_in_vec_max, 0, NULL, &event);
		status = clWaitForEvents(1, &event);
		clReleaseEvent(event);
		err_check(errNum, "clSetMemObj");

		while ((az_file[0] > id_min) && (az_file[0] <= id_max)) {
			ofstream visible_fp;
			if (flag_tmstmp == 1) {

				stringstream ss;
				string str;
				ss
						<< "/users/figu_ra/workspace/Illu_CL/results/visible_files/visible_"
						<< fixed << eft[i_count] << ".txt";
				ss >> str;
				visible_fp.open(str.data());
			}

			//---- HERE STARTS KERNEL


			mobj_az_file = clCreateBuffer(context, CL_MEM_READ_WRITE,
					sizeof(float) * 1, NULL, NULL);
			errNum = clEnqueueWriteBuffer(command_queue, mobj_az_file, CL_TRUE,
					0, sizeof(float) * 1, az_file, 0, NULL, &event);
			status = clWaitForEvents(1, &event);
			clReleaseEvent(event);
			err_check(errNum, "clSetMemObj");

			mobj_sun = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float)
					* 5, NULL, NULL);
			errNum = clEnqueueWriteBuffer(command_queue, mobj_sun, CL_TRUE, 0,
					sizeof(float) * 5, sun_file, 0, NULL, &event);
			status = clWaitForEvents(1, &event);
			clReleaseEvent(event);
			err_check(errNum, "clSetMemObj");

			mobj_out = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float)
					* size_roi, NULL, NULL);

			// Step 10: Set OpenCL kernel argument

			errNum = clSetKernelArg(kernel, 0, sizeof(cl_mem), &mobj_x_vec);
			errNum = clSetKernelArg(kernel, 1, sizeof(cl_mem), &mobj_y_vec);
			errNum = clSetKernelArg(kernel, 2, sizeof(cl_mem), &mobj_z_vec);

			errNum = clSetKernelArg(kernel, 3, sizeof(cl_mem), &mobj_elev_min);
			errNum = clSetKernelArg(kernel, 4, sizeof(cl_mem), &mobj_elev_max);

			errNum = clSetKernelArg(kernel, 5, sizeof(cl_mem), &mobj_az_file);
			errNum = clSetKernelArg(kernel, 6, sizeof(cl_mem), &mobj_sun);
			errNum = clSetKernelArg(kernel, 7, sizeof(cl_mem), &mobj_out);

			if (errNum != CL_SUCCESS) {
				cerr << "Error queuing kernel for execution." << endl;
				return 1;
			}

			errNum = clEnqueueNDRangeKernel(command_queue, kernel, 1, NULL,
					globalWorkSize, localWorkSize, 0, NULL, NULL);
			if (errNum != CL_SUCCESS) {
				cerr << "Error queuing kernel for execution." << endl;
				return 1;
			}

			clFinish(command_queue);
			errNum = clEnqueueReadBuffer(command_queue, mobj_out, CL_TRUE, 0,
					sizeof(float) * size_roi, visible, 0, NULL, &event);
			status = clWaitForEvents(1, &event);
			clReleaseEvent(event);

			//-----WRITE OUT THE FILE

			for (int i = 0; i < size_roi; i++) {

				if (flag_tmstmp == 1) {
					visible_fp << visible[i] << endl;
				}

				//if (visible[i] < 0.001) {
				//	visible[i] = 0;
				//}
				if ((visible[i] != visible[i]) || (visible[i] < 0)) {
					visible[i] = 0;
				}

				accum_temp[i] = accum_temp[i] + visible[i];
				if (i == lqcip) {//7322

					sunsun_fp << fixed << setprecision(8) << order_vec[i_count]
							<< " " << visible[i] << endl;

				}

			}

			//------ HERE ENDS KERNEL


			clReleaseMemObject(mobj_az_file);
			clReleaseMemObject(mobj_sun);
			clReleaseMemObject(mobj_out);
			clFinish(command_queue);
			clFlush(command_queue);
			//need the write out part
			if (flag_tmstmp == 1) {
				visible_fp.close();
			}

			i_count++;
			az_file[0] = az_vec[i_count];

			sun_file[0] = sun_x_v[i_count];
			sun_file[1] = sun_y_v[i_count];
			sun_file[2] = sun_z_v[i_count];

			off++;
		}
		clReleaseMemObject(mobj_elev_max);
		clReleaseMemObject(mobj_elev_min);
		id_min = az_in_vec[read_count + 1];
		sun_file[3] = id_min;

		if (read_count == 0) {
			cout << "[                    ]0%\r" << flush;
		} else if (read_count == 144) {
			cout << "[====                ]20%\r" << flush;

		} else if (read_count == 288) {
			cout << "[========            ]40%\r" << flush;

		} else if (read_count == 432) {
			cout << "[============        ]60%\r" << flush;

		} else if (read_count == 576) {
			cout << "[================    ]80%\r" << flush;

		}

	}
	readin[0].close();
	sunsun_fp.close();

	int year = out / 8760;
	cout << "[====================]100%\r" << endl;
	cout << "Calculation of illumination finished..." << endl;

	cout << " Writing out results..." << endl;
	ofstream accum_per, accum_days;
	accum_days.open(
			"/users/figu_ra/workspace/Illu_CL/results/accumulated_days.txt");
	accum_per.open(
			"/users/figu_ra/workspace/Illu_CL/results/accumulated_percentage.txt");
	for (int i = 0; i < size_roi; i++) {
		accum_temp[i] = accum_temp[i] / (24 * year);
		accum_days << accum_temp[i] << endl;
		accum_temp[i] = (accum_temp[i] * 24 * year) / out;
		accum_per << fixed << setprecision(6) << accum_temp[i] << endl;
	}
	accum_days.close();
	accum_per.close();

	delete[] az_vec;
	delete[] order_vec;
	delete[] sun_x_v;
	delete[] sun_y_v;
	delete[] sun_z_v;
	delete[] eft;
	delete[] elev_in_vec_min;
	delete[] elev_in_vec_max;
	delete[] x_vec;
	delete[] y_vec;
	delete[] z_vec;
	delete[] visible;
	delete[] accum_temp;

	clReleaseKernel(kernel);
	clReleaseCommandQueue(command_queue);
	clReleaseProgram(program);
	clReleaseContext(context);

	time = clock() - time;
	double ms = double(time) / CLOCKS_PER_SEC;
	cout << "Elapsed time: " << ms / 60 << " minutes" << endl;
	cout << "Finished succesfully." << endl;

}
